package edu.mit.ll.group43.surfaceoptimization.dp.factory.cost;

import java.util.ArrayList;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.WeightClass;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkEdge;
import edu.mit.ll.group43.surfaceoptimization.network.dp.CPSNetworkNode;

/**
 * The spot cost strategy for the All Gaps algorithm. This algorithm considers
 * each gap in the projected takeoff sequence when scheduling the current aircraft,
 * only taking into account gaps that are already large enough to fit the current
 * aircraft without having to reschedule previously solved for aircraft. 
 * 
 * @author William Hawkins
 *
 */
public class AllGapsSpotCostStrategy extends OneGapSpotCostStrategy {

	public AllGapsSpotCostStrategy(int cps,
			MinimumSeparationCalculator<? extends WeightClass> calc) {
		super(cps, calc);
	}
	
	/**
	 * Returns the "predecessor bound" for the current aircraft.
	 * 
	 * @param prevNode
	 * @param current
	 * @return
	 */
	public int safePrevTimeBound(CPSNetworkNode prevNode, Aircraft current) {
		return prevNode.getLastAircraft().getOptReleaseTime();
	}
	
	/**
	 * Determines the earliest time at which the last aircraft of the target node
	 * of the given network edge can be released from the spot. Considers all gaps
	 * between takeoffs in the given takeoff sequence and tries to fit the current
	 * aircraft into the earliest one it can. Failing that, it sends it as early
	 * as it can to minimize how long it takes off after the last projected takeoff
	 * time, while respecting the optimal spot release time of this aircraft's
	 * direct predecessor in the CPS spot release sequence.
	 * 
	 * @param edge
	 * @param sortedTakeOffs
	 */
	public int earliestRelease(NetworkEdge edge, ArrayList<Aircraft> sortedTakeOffs) {
		Aircraft lastReleasedAircraft = ((CPSNetworkNode) edge.getTarget()).getLastAircraft();
		CPSNetworkNode optPrevNode = (CPSNetworkNode) edge.getSource();
		
		for (int i = 0; i < sortedTakeOffs.size(); i ++) {
			Aircraft firstReadyAircraft = null;
			if (sortedTakeOffs.size() >= 2 && i >= 1) {
				firstReadyAircraft = sortedTakeOffs.get(i-1);
			}
			Aircraft secondReadyAircraft = sortedTakeOffs.get(i);
			
			// Determine the earliest that this aircraft can be released while satisfying
			// all constraints
			int earlyRelease = earliestReleaseBetween(safePrevTimeBound(optPrevNode, lastReleasedAircraft),
					firstReadyAircraft, secondReadyAircraft, lastReleasedAircraft);
			if (earlyRelease > -1) {
				return earlyRelease;
			}
		}
		
		return earliestRelease(sortedTakeOffs.get(sortedTakeOffs.size()-1), lastReleasedAircraft,
				safePrevTimeBound(optPrevNode, lastReleasedAircraft));
	}
	
	/**
	 * Determines the earliest time at which the last aircraft of the target node
	 * of the given network edge can be released from the spot. Considers all gaps
	 * between takeoffs in the current projected takeoff sequence and tries to fit
	 * the current aircraft into the earliest one it can. Failing that, it sends it
	 * as early as it can to minimize how long it takes off after the last projected
	 * takeoff time, while respecting the optimal spot release time of this aircraft's
	 * direct predecessor in the CPS spot release sequence.
	 * 
	 */
	@Override
	public int earliestRelease(NetworkEdge edge) {
		return earliestRelease(edge, sortedAlphasSoFar(edge));
	}

}
