package edu.mit.ll.group43.surfaceoptimization.dp.factory.cost;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.TreeSet;

import edu.mit.ll.group43.surfaceoptimization.aircraft.Aircraft;
import edu.mit.ll.group43.surfaceoptimization.aircraft.AircraftAlphaComparator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.AircraftSpotReleaseComparator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator;
import edu.mit.ll.group43.surfaceoptimization.aircraft.MinimumSeparationCalculator.ActionType;
import edu.mit.ll.group43.surfaceoptimization.aircraft.WeightClass;
import edu.mit.ll.group43.surfaceoptimization.network.NetworkEdge;
import edu.mit.ll.group43.surfaceoptimization.network.dp.CPSNetworkNode;
import edu.mit.ll.group43.surfaceoptimization.network.dp.SmartCPSNetworkNode;
import edu.mit.ll.group43.surfaceoptimization.validator.CPSValidator;

/**
 * Spot cost strategy for the One Gap Force algorithm. This algorithm
 * considers the last gap in the projected take off sequence, and determines
 * whether forcing this gap or scheduling the current aircraft to take off
 * last incurs the least additional delay.
 * 
 * @author William Hawkins
 *
 */
public class OneGapForceSpotCostStrategy extends OneGapSpotCostStrategy {

	protected HashMap<CPSNetworkNode, Integer> optSpotReleases;
	
	public OneGapForceSpotCostStrategy(int cps,
			MinimumSeparationCalculator<? extends WeightClass> calc) {
		super(cps, calc);
		optSpotReleases = new HashMap<CPSNetworkNode, Integer>();
	}
	
	/**
	 * Sets the optimal values for the node being solved for as well as its
	 * last aircraft.
	 * 
	 * @param edge
	 * @param distance
	 */
	@Override
	public void setOptimalValues(NetworkEdge edge, int distance) {
		SmartCPSNetworkNode prev = (SmartCPSNetworkNode) edge.getSource();
		SmartCPSNetworkNode node = (SmartCPSNetworkNode) edge.getTarget();
		Aircraft follower = node.getLastAircraft();
		
		// Set the optimal distance, prev node, and spot release time
		node.setOptDistance(distance);
		node.setOptPrevNode(edge);
		follower.setOptReleaseTime(optSpotReleases.get(prev));

		// Initialize this node's release sequence with that of the
		// previous node
		ArrayList<Aircraft> nodeReleaseSequence = new ArrayList<Aircraft>();
		for (Aircraft ac : prev.getReleaseSequence()) {
			nodeReleaseSequence.add(ac.clone());
		}
		
		// Calculate the new spot release time of the last aircraft to
		// take off and set it (only in the sequence for this node) if
		// necessary
		int addedDistance = distance - prev.getOptDistance();
		int diff = addedDistance - (optSpotReleases.get(prev) - follower.getSpotReadyTime());
		Aircraft lastTakeOff = null;
		if (diff > 0) {
			Collections.sort(nodeReleaseSequence, AircraftAlphaComparator.getInstance());
			lastTakeOff = nodeReleaseSequence.get(nodeReleaseSequence.size()-1);
			lastTakeOff.setOptReleaseTime(lastTakeOff.getOptReleaseTime() + diff);
		}
		
		// Add the current aircraft to the sequence, sort according to
		// optimal spot release time, and then assign it to the node.
		nodeReleaseSequence.add(follower);
		Collections.sort(nodeReleaseSequence, AircraftSpotReleaseComparator.getInstance());
		node.setReleaseSequence(nodeReleaseSequence);
		
		// Clear the release time mappings for the next node to use
		// this strategy
		optSpotReleases.clear();
	}
	
	/**
	 * Returns a list of aircraft cloned from the release sequence of the source node
	 * of the given edge, sorted by runway ready time.
	 * 
	 * @param edge
	 */
	@Override
	public ArrayList<Aircraft> sortedAlphasSoFar(NetworkEdge edge) {
		ArrayList<Aircraft> releaseSequence = ((SmartCPSNetworkNode) edge.getSource()).getReleaseSequence();
		return sortedAlphasSoFar(releaseSequence);
	}
	
	/**
	 * Returns a list of aircraft cloned from the given release sequence, sorted by
	 * runway ready time.
	 * 
	 * @param sequence
	 * @return
	 */
	public ArrayList<Aircraft> sortedAlphasSoFar(ArrayList<Aircraft> sequence) {
		ArrayList<Aircraft> alphaSequence = new ArrayList<Aircraft>();
		for (Aircraft ac : sequence) {
			alphaSequence.add(ac.clone());
		}
		Collections.sort(alphaSequence, AircraftAlphaComparator.getInstance());
		return alphaSequence;
	}
	
	/**
	 * Computes the earliest time at which current can be released, with respect
	 * to its release sequence predecessors in the spot release sequence of the
	 * given previous node.
	 * 
	 * WARNING: Should always be compared with current's spot ready time since
	 * this method will return zero if none of the aircraft already placed in
	 * the release sequence can follow current.
	 * 
	 * @param prevNode
	 * @param current
	 * @return
	 */
	public int safePrevTimeBound(CPSNetworkNode prevNode, Aircraft current) {
		return safePrevTimeBound(((SmartCPSNetworkNode) prevNode).getReleaseSequence(), current);
	}
	
	/**
	 * Computes the earliest time at which current can be released, with respect
	 * to its release sequence predecessors.
	 * 
	 * WARNING: Should always be compared with current's spot ready time since
	 * this method will return zero if none of the aircraft already placed in
	 * the release sequence can follow current.
	 * 
	 * @param releaseSequence
	 * @param current
	 * @return
	 */
	public int safePrevTimeBound(ArrayList<Aircraft> releaseSequence, Aircraft current) {
		return CostStrategyUtils.safePrevTimeBound(releaseSequence, current, CPS);
	}
	
	/**
	 * Initializes the first stage of the CPS network so that nodes in subsequent stages
	 * can be solved for.
	 * 
	 * @param network
	 */
	@Override
	public void initializeFirstStage(ArrayList<TreeSet<CPSNetworkNode>> network) {
		super.initializeFirstStage(network);
		for (CPSNetworkNode node : network.get(0)) {
			((SmartCPSNetworkNode) node).getReleaseSequence().add(node.getLastAircraft());
		}
	}
	
	/**
	 * Returns the additional delay associated with the given edge that is given by the
	 * One Gap Force algorithm.
	 * 
	 * @param edge
	 */
	@Override
	public int getEdgeWeight(NetworkEdge edge) {
		
		// Reference to the source node
		CPSNetworkNode optPrevNode = (CPSNetworkNode) edge.getSource();
		
		// Get the last two aircraft projected to take off
		ArrayList<Aircraft> takeOffs = sortedAlphasSoFar(edge);
		Aircraft b = takeOffs.get(takeOffs.size()-1);
		Aircraft a = null;
		if (takeOffs.size() >= 2) {
			a = takeOffs.get(takeOffs.size()-2);
		}
		
		// Set the earliest release to what lines up with the earliest the
		// follower can follow aircraft a, disregarding aircraft b
		Aircraft follower = ((CPSNetworkNode) edge.getTarget()).getLastAircraft().clone();
		int earliestRelease = earliestReleaseAfter(safePrevTimeBound(optPrevNode, follower), a, follower);
		
		// Determine how this affects the last aircraft. Aircraft b will need
		// to make its optimal spot release time later than it is by the number
		// of seconds of overlap between follower's takeoff and wake vortex
		// separation and b's runway ready time. This difference will add delay.
		int diff = earliestRelease + follower.getUnimpTaxiTime() + 
				CALC.getSeparation(follower.getWeightClass(), ActionType.DEPARTURE, 
						b.getWeightClass(), ActionType.DEPARTURE) - b.getRunwayReadyTime();
		
		// If they do overlap, hold b's optimal release time back by that amount
		// so that the effects can be measured below when the list is sorted
		if (diff < 0) {
			diff = 0;
		}
		
		// The additional delay for the forcing method is the delay added to b
		// plus the delay for follower.
		int gapForceDelay = diff + earliestRelease - follower.getSpotReadyTime();
		
		// Get what would happen otherwise
		int normalRelease = earliestRelease(b, follower, safePrevTimeBound(optPrevNode, follower));
		int normalDelay = normalRelease - follower.getSpotReadyTime();
		
		if (gapForceDelay <= normalDelay) {
			Collections.sort(takeOffs, AircraftSpotReleaseComparator.getInstance());
			int origBIndex = takeOffs.indexOf(b);
			b.setOptReleaseTime(b.getOptReleaseTime() + diff);
			follower.setOptReleaseTime(earliestRelease);
			takeOffs.add(follower);
			if (CPSValidator.validate(takeOffs, takeOffs.get(Math.max(0, origBIndex-1)), CPS)) {
				optSpotReleases.put(optPrevNode, earliestRelease);
				return gapForceDelay;
			}
		}
		optSpotReleases.put(optPrevNode, normalRelease);
		return normalDelay;
	}

}
